"""
    easy_graphql_server is an interface to easily expose a database in GraphQL via ORM models.
"""

from .types import JSONString, Required, Model
from .schema import Schema
from .operations import Operation
from .exceptions import UnauthenticatedError, NotFoundError, ForbiddenError, \
    ValidationError, DuplicateError, IntegrityError
from .exposition import ExposedModel, ExposedQuery, ExposedMutation, CustomField

CREATE = Operation.CREATE
READ = Operation.READ
UPDATE = Operation.UPDATE
DELETE = Operation.DELETE
